#!/usr/local/bin/perl
#############################################################################
# FW-1 log summarizer by rui.bernardino@necdum.net
# copyright (c) 1999 Rui Bernardino
#               2000 Paul J. Ewing Jr.
#
# CHANGES
# 0.2.2: More bug fixes: added 'undef' for all variables in the main loop.
#                        a few english corrections.
#                        some reports were broken (like DNS tranfers)
#        
# 0.2.1: Reports dont use colons as separator (Perry Morgan)
# 0.2.0: Too many fixes/enhancements
# 0.1.0: Initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

#
# For bug reports and suggestions contact me at rui.bernardino@necdum.net
#

# defaults command line options are set here
%opctl = (
    "outputdir"     => ".", # current directory
    "maxtable"      => 0,   # max table rows, doesn't override
                            # per report maximum (if set). default is none.
    "hostresolv"    => 0,   # do not try to resolve unknow IPs, slow.
                            # Doesn't override fw-1 definitions
    "serviceresolv" => 1,   # do not try to resolve unknow services, slow.
                            # Doesn't override fw-1 definitions
    "dnscache"      => 1,   # Cache DNS querie results.
    "showunmatched" => 1,   # Show unmatched lines
    "method"        => 0,   # Autodetect method
    "conf"          => "./fwlogsum.conf"
                            # the default configuration file is the current
                            # directorie
);


# parse options
if(!GetOptions(\%opctl,
		"maxtable=i",
		"method=i",
		"outputdir=s",
		"conf=s",
		"hostresolv!",
		"dnscache!",
		"showunmatched!",
		"serviceresolv!",
		"help!")
	||$opctl{"help"}) {
	print <<USAGE;
usage: ./fwlogsum [options...] [logfiles]
Available options are:
--outputdir=dir         directory for HTML report output (default is "current")
--conf=file             configuration file (default is "./fwlogsum.conf")
--method=num            1 for "export", 2 for "syslog", default autodetect.
--[no]showunmatched		Show unmatched log lines. Default if on.
--maxtable=num          max table rows, does not override per report maximum
                        (if set). default is none.
--[no]hostresolv        try to resolve unknow IPs, slow. Doesnt override fw-1
                        definitions. default is no.
--[no]serviceresolv     try to resolve unknow services, slow. Doesnt override
                        fw-1 definitions. default is no.
--[no]dnscache			disable internal hack to cache DNS. Useless if you have
						a decent DNS cache or insuficient memory. Default on.
--help                  This screen.
Input can be any number of files and/or standard input.

USAGE
	exit(-1);
}

# print out options
#foreach (keys %opctl) {print "$_ $opctl{$_}\n";}

# open configuration file
require $opctl{conf};

$outputlines=0;
use Getopt::Long;
use Net::hostent;
use Net::servent;

# This are the basis of the parsing engine. I hope you don't need to mess with these.
# If you have any suggestions for a more generic (and fast) parser, let me know
$sysloghead='(\w+)\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+[\d\:]+\s+(\S+)\s+(\S+)\s+[<>]([^<>\s]+)(.*$)'; 

#$exporthead='^num;date;time;orig;type;action;alert;i/f_name;i/f_dir;proto;src;dst;service;s_port;len;rule';
$exporthead='^\d+;\d+\w+\d+;\s*\d+:\d+:\d+\;\S+;\w+;\w+;\w*;\S+;\w+;\w+;\S+;\S+;\S+;\w+;\d+;\d+';

$method=$opctl{"method"};

$UNKNOWN = "unknown";

# This is a mess.... maybe I should ler perl
while(<>) {
	# ignore blank lines
	next if /^[\n\r\s]*$/;
	# undef all previous values, just in case
	undef $date; undef $time; undef $firewall;
	undef $op; undef $interface;
        undef $proto; undef $src; undef $dst;
	undef $service; undef $s_port; undef $len; undef $rule;
	undef $will_see; undef $xlatedst; undef $xlatesport;
	undef $xlatedport; undef $icmptype; undef $icmpcode; undef $sysmsgs;
	if(!$method) {
			# if this is export, we should skip the first line
		if(/$exporthead/) { $method=1;next;}
			# but not if it is syslog
		elsif(/$sysloghead/) { $method=2;}
			# try again...
		else { print "Coulnd't detect method\n";next; }

	}
	# yes, i'm working
	print "$.\n" if $.%1000 == 0;

	# set initial date and parse input line
	# this code really need improvement, but lots of care must be taken
	# to avoid huge performance impact
	if($method==1) {
			# spliting the lines will speed up things... alot
			($nil,$date,$time,$firewall,$nil,$op,$nil,$interface,$nil,
			$proto,$src,$dst,$service,$s_port,$len,$rule,$will_see,
			$xlatedst,$xlatesport,$xlatedport,$icmptype,$icmpcode,$sysmsgs)=split /\;/;
			# unfortunately, checkpoint consider "daemon" to be a interface name.
			#if($interface eq "daemon") {next;}
			if(!$sysmsgs) {$sysmsgs=$will_see;}
			else {$xlatesrc=$will_see;}
			# strip white spaces
			$time=~s/\s*//g;
			if($sysmsgs) {
				# remove carriage returns and commas
				$sysmsgs=~s/[\n\r\,]//g;
				$sysmsgs=~s/^\s*(\w.*?\w)\s*$/$1/;
				#if($sysmsgs) {print "rest=[(",$sysmsgs,")]\n";}
			}
			($day,$month)=($date=~m/^(\d+)(\D+)/);
				#print "Method $method matched\n$_\n";
		}
	# this 
	elsif ($method==2&& (($month,$day,$time,$op,$firewall,$interface,$sysmsgs)=m/$sysloghead/)) {
				$proto=$1 if $sysmsgs=~s/proto (\S+)//;
				$src=$1 if $sysmsgs=~s/src (\S+)//;
				$dst=$1 if $sysmsgs=~s/dst (\S+)//;
				$service=$1 if $sysmsgs=~s/service (\S+)//;
				$s_port=$1 if $sysmsgs=~s/s_port (\S+)//;
				$len=$1 if $sysmsgs=~s/len (\S+)//;
				$rule=$1 if $sysmsgs=~s/rule (\S+)//;
				$xlatesrc=$1 if $sysmsgs=~s/xlatesrc (\S+)//;
				$xlatedst=$1 if $sysmsgs=~s/xlatedst (\S+)//;
				$xlatesport=$1 if $sysmsgs=~s/xlatesport (\S+)//;
				$xlatedport=$1 if $sysmsgs=~s/xlatedport (\S+)//;
				$icmptype=$1 if $sysmsgs=~s/icmp-type (\S+)//;
				$icmpcode=$1 if $sysmsgs=~s/icmp-code (\S+)//;
				#print "$proto $src $dst $service $s_port $len $rule $xlatesrc $xlatedst $xlatesport $xlatedport\n";
				# remove carriage returns and commas
				$sysmsgs=~s/[\n\r\,]//g;
				# strip white space left by previous replaces
				undef $sysmsgs unless $sysmsgs=~s/^\s*(\w.*?\w)\s*$/$1/;
				#if($sysmsgs) {print "rest=[(",$sysmsgs,")]\n";}
				#print "Method $method matched\n$_\n";
		}
	# Failed to match anything
	else {
		if ($opctl{'showunmatched'} ) {print "NO MATCH (method $method)\n$_";}
		next;
	}

	if(!$im) { ($im,$id,$ih)=($month,$day,$time);}
	($hour,$minute,$second)=split /\:/, $time;
	$hour=~s/^\s*(\d)$/0$1/;
	$minute=~s/^\s*(\d)$/0$1/;
	$second=~s/^\s*(\d)$/0$1/;

	# store last successfuly parsed time
	$lasttime="$time $day $month" if $time;

	$outputlines++;
	# resolv service numbers to service names
	if($opctl{"serviceresolv"}) {
		foreach (\$service,\$port) {
			# service names *should not* start with digits
			if($$_=~/^\d/) {
				# keep a services cache
				if(!$servicecache{$$_}) {
					$s=getserv($$_);
					if($s) {
						$servicecache{$$_}=$s->name;
					}
					else {
						 $servicecache{$$_}="($$_)";
					}
				}
				$$_=$servicecache{$$_};
			}
		}
	}

	# core report engine
	foreach $r (keys %reports) {
		if(&{$reports{$r}[0]}) {
			# determine the major key for the new hash. Drop if any of the
			# needed variables is null
			@key = map { ${$_} || next } @{$reports{$r}[1]};
			$key = join ",", $r, @key;
			$packets{$key}++;
			$bytes{$key}+=$len;
			if($r=~m/^DEBUG/) {
				print "$r: stored $packets{$key}: $bytes{$key}.  matched at $hour:$minute:$second ($op,$firewall,$interface,$proto,$src,$dst,$service, $s_port,$len,$rule)\n";
			}
		}
		else {
			if($r=~m/^NDEBUG/) {
				print "$r failed in ($op,$firewall,$interface,$proto,$src,$dst,$service, $s_port,$len,$rule)\n";
			}
		}

	}
}


# open output file
# HTML tweaking here
$end=join "-",split /\s/,$lasttime;
$start="$ih-$id-$im";

if ($start&&$end) {
	print "last=$lasttime hdm=$ih $id $im start=$start end=$end\n";
	$fls="fls-$start-$end.html";
	# remove those nasty colons that break windows compatibility
	$fls=~s/\:/_/g;
	open OUT,">$opctl{\"outputdir\"}/$fls" or die $!;

	print OUT "<html><head>\n<title>Firewall Summary between $ih $id $im",
		" and $lasttime</title>\n</head><body bgcolor=\"#FFFFFF\">\n",
		"<h1><a NAME=\"#DOCUMENT_START\">Firewall Summary between $ih $id $im and $lasttime</a></h1>\n",

	# control information
		"<!-- FWLOGSUM_MARK (Firewall Summary between $ih $id $im and $lasttime) ($fls) -->\n";

	# print TOC
	print OUT "</br><h2>Table of Contents</h2>\n";
	foreach $r (sort keys %reports) {
			($label)=split /\,/,$reports{$r}[2];
			print OUT "<a href=\"#$r\">$label</a><br>\n";
			}


	# report output engine
	# awfull mess
	foreach $p (sort order keys %packets) {
		# one of these days, I really should try to use hashes of hashes of hashes...
		($head,$cols)=split ",",$p,2;
		# If this is true, than we somehow changed to a new report rule
		if($head ne $last) {
			close_table();
			}
		# only print these lines if they fit max configured size
		if($head=~m/^DEBUG/) {
			print "Output: $head -> $packets{$p},$bytes{$p}\n";
			}
		if((!$maxtablelines&&(!$opctl{"maxtable"}||$opctl{"maxtable"}>$lines))
		||$maxtablelines>$lines) {
			$lines++;
			print OUT "<tr>\n";
			foreach $col (split ",",$cols) {
				if($opctl{"hostresolv"} && $col =~ /^\d+\./) {
					gethostname(\$col);
				}
				print OUT "<td>$col</td>\n";
				}
			print OUT "<td align=right>$packets{$p}</td>",
				"<td align=right>$bytes{$p}</tr>\n";
			
		} #else {print "No more after $lines for maximum $maxtablelines for report $head\n";}
		# Only store graph values as required
		if(!$maxgraphlines || $maxgraphlines>$glines) {
			$glines++;
			$gvalues{$cols}=$packets{$p};
			}
	}
	
	close_table(); 
			
	print OUT "</table></center><p>\n";
	
	print OUT "</body></html>";
	print "done. $outputlines parsed.\n";
	
	close OUT;
} else {
	print "\nNot enough valid log entries were found!\n";
	}
	
# and now, generate the prettiest index in the world! And unsorted!
@titles=`head -10 \`ls -tr $opctl{outputdir}\/fls*.html\` | grep FWLOGSUM_MARK`;

open INDEX,">$opctl{outputdir}/index.html" or die $!;

print INDEX "<html><head>\n<title>Firewall Summaries\n",
	"Firewall Summaries</title>\n</head>\n",
	"<h1>Firewall Summaries</h1>\n<body>\n";

foreach (@titles) {
	# strip html marks
	if(/\((.*)\) \((.*)\)/) {
			print INDEX "<A HREF=\"$2\">$1</a><p>\n";
		}
	};
print INDEX '</body></html>';
close INDEX;

# *try* to copy the texture into the output directory
system("cp $texture $opctl{outputdir}");




# sort by number of conections
sub order {
	($ha)=split ",",$a;
	($hb)=split ",",$b;
	if($ha eq $hb) {
		return ($packets{$b} <=> $packets{$a});
		}
	else {
		return ($ha cmp $hb);
	}
}


# one more dirty trick, if this was BASIC, it would be a subrotine ;)
sub close_table {
		# if this is not the first table
		if($last) {
			print OUT "</table></td></tr>\n";

			if($glines) {
				# Now it is time to print out the histogram
				print OUT "<tr><td align=left>";
				# find out max and min values
				undef $max, $min;
				foreach $g (keys %gvalues) {
					if(!$max || $gvalues{$g}>$gvalues{$max}) {$max=$g;}
					if(!$min || $gvalues{$g}<$gvalues{$min}) {$min=$g;}
					}
				# calculate q, the rude way
				$q=$max_graph_pixels/$gvalues{$max};
				# output all the values
				print OUT "<table cellpadding=0 cellspacing=0 width=100% border=0>\n";
				# this is weird
				@repeated=();
				foreach $g (sort keys %gvalues) {
					# the label
					@label=split /\,/,$g;
					foreach (0.. $#label ) {
						if($opctl{"hostresolv"} && $label[$_]=~ /^\d+\./) {
							gethostname($label[$_]);
						}
						$label[$_]="" if $label[$_] eq $repeated[$_] or last;
						}
					print OUT "<tr align=left><td align=right><b>".join(" ",@label)."</b></td>\n";
					# the bar
					print OUT "<td align=left><table><tr><td><img border=0 src='".$texture."' height=16 width=",
								int($gvalues{$g}*$q),"></td><td>$gvalues{$g}</td></tr></table></td></tr>\n";
					@repeated=split /\,/,$g;
					}
				print OUT "</table></td></tr>";
				undef %gvalues;
				}
			print OUT "</table></center>";
			}

		
		# get the information from %reports
		if($head&&$head ne $last) {
			($section,$colname)=split ",",$reports{$head}[2],2;
			$maxtablelines=$reports{$head}[3];
			$maxgraphlines=$reports{$head}[4];
			$lines=$glines=0;
	
			print OUT "</center><br><br><br>\n\n\n<a NAME=\"$head\"></a><h2>$section ";
			if($reports{$head}[3]>0) {print OUT "(top $reports{$head}[3])";}
			#print OUT "</h2><a href=\"#DOCUMENT_START\" NAME=\"$head\">Up</a><center>";
			print OUT "</h2><a href=\"#DOCUMENT_START\">Up</a><center>";
			print OUT "<table border=0 NOSAVE>\n";
			if($maxtablelines>=0) {
				print OUT "<tr><td><table width=100% border=1><tr>\n";
				#print "HEAD: $head \"$section\" $colname $maxtablelines $maxgraphlines $lines $glines\n";
				foreach(split ",",$colname) {
					print OUT "<td><b>$_</b></td>";
				}
				print OUT "<td><b>events</b></td><td><b>bytes</b></td></tr>\n\n";
			}
			$last=$head;
			}
}

sub gethostname {
    my(@hosts) = @_;

    foreach my $host (@hosts) {
	# keep a host cache
	if(!$hostcache{$$host}) {
	    $h=gethost($$host);
	    if($h) {
		$hostcache{$$host}=$h->name;
	    }
	    else {
		$hostcache{$$host}="($$host)";
	    }
	}
	$$host=$hostcache{$$host};
	if(!$opctl{"dnscache"}) {
	    # ugly but simple
	    undef $hostcache{$$host};
	}
    }
}
